package fit;

// Copyright (c) 2002 Cunningham & Cunningham, Inc.
// Released under the terms of the GNU General Public License version 2 or later.

import java.io.*;
import java.util.*;
import java.lang.reflect.*;
import java.text.DateFormat;

public class Fixture {
    static int right = 0;
    static int wrong = 0;
    static int ignores = 0;
    static int exceptions = 0;
    static Map summary = new HashMap();


    // Traversal //////////////////////////

    public void doTables(Parse tables) {
        summary.put("run date", new Date());
        while (tables != null) {
            Parse heading = tables.at(0,0,0);
            if (heading != null) {
                try {
                    Fixture fixture = (Fixture)(Class.forName(heading.text()).newInstance());
                    fixture.doTable(tables);
                } catch (Exception e) {
                    exception (heading, e);
                }
            }
            tables = tables.more;
        }
    }

    public void doTable(Parse table) {
        doRows(table.parts.more);
    }

    public void doRows(Parse rows) {
        while (rows != null) {
            doRow(rows);
            rows = rows.more;
        }
    }

    public void doRow(Parse row) {
        doCells(row.parts);
    }

    public void doCells(Parse cells) {
        for (int i=0; cells != null; i++) {
            try {
                doCell(cells, i);
            } catch (Exception e) {
                exception(cells, e);
            }
            cells=cells.more;
        }
    }

    public void doCell(Parse cell, int columnNumber) {
        ignore(cell);
    }


    // Annotation ///////////////////////////////

    public void right (Parse cell) {
		cell.addToTag(" bgcolor=\"#cfffcf\"");
		right++;
	}

    public void wrong (Parse cell) {
        cell.addToTag(" bgcolor=\"#ffcfcf\"");
        wrong++;
    }

    public void wrong (Parse cell, String actual) {
        wrong(cell);
        cell.addToBody(label("expected") + "<hr>" + actual + label("actual"));
    }

    public void ignore (Parse cell) {
        cell.addToTag(" bgcolor=\"#efefef\"");
        ignores++;
    }

    public void exception (Parse cell, Exception exception) {
        if(exception.getClass().equals(InvocationTargetException.class)) {
            exception = (Exception)((InvocationTargetException)exception).getTargetException();
        }
        final StringWriter buf = new StringWriter();
        exception.printStackTrace(new PrintWriter(buf));
        cell.addToBody("<hr><font size=-2><pre>" + (buf.toString()) + "</pre></font>");
        cell.addToTag(" bgcolor=\"#ffffcf\"");
        exceptions++;
    }

    // Utility //////////////////////////////////

    public static String counts() {
        return
            right + " right, " +
            wrong + " wrong, " +
            ignores + " ignored, " +
            exceptions + " exceptions";
    }

    static String label (String string) {
        return " <font size=-1 color=#c08080><i>" + string + "</i></font>";
    }

    static String camel (String name) {
        StringBuffer b = new StringBuffer(name.length());
        StringTokenizer t = new StringTokenizer(name);
        b.append(t.nextToken());
        while (t.hasMoreTokens()) {
            String token = t.nextToken();
            b.append(token.substring(0, 1).toUpperCase());      // replace spaces with camelCase
            b.append(token.substring(1));
        }
        return b.toString();
    }

    public Object parse (String s, Class type) throws Exception {
        if (type.equals(String.class))  {return s;}
        if (type.equals(Date.class))    {return DateFormat.getDateInstance().parse(s);}
        throw new Exception("can't yet parse "+type);
    }

    void check(Parse cell, TypeAdapter a) {
        String text = cell.text();
        if (text.equals("")) {
            return;
        }
        if (text.equals("error")) {
            try {
                Object result = a.invoke();
                wrong(cell, a.toString(result));
            } catch (IllegalAccessException e) {
                exception (cell, e);
            } catch (Exception e) {
                right(cell);
            }
            return;
        }
        try {
            Object result = a.get();
            if (a.equals(a.parse(text), result)) {
                right(cell);
            } else {
                wrong(cell, a.toString(result));
            }
        } catch (Exception e) {
            exception(cell, e);
        }
    }

}
