// Copyright (c) 2002 Cunningham & Cunningham, Inc.
// Released under the terms of the GNU General Public License version 2 or later.

package fit;

import java.util.*;

abstract public class RowFixture extends ColumnFixture {

    Object results[];
    List missing = new LinkedList();
    List surplus = new LinkedList();


    public void doRows(Parse rows) {
        bind(rows.parts);
        results = query();
        match(list(rows.more), list(results), 0);
        Parse last = rows.last();
        last.more = buildRows(surplus.toArray());
        mark(last.more, "surplus");
        mark(missing.iterator(), "missing");
    }

    abstract public Object[] query();           // get rows to be compared
    abstract public Class getTargetClass();     // get expected type of row

    void match(List expected, List computed, int col) {
        Map eMap = eSort(expected, col);
        Map cMap = cSort(computed, col);
        Set keys = union(eMap.keySet(),cMap.keySet());
        for (Iterator i=keys.iterator(); i.hasNext(); ) {
            Object key = i.next();
            List eList = (List)eMap.get(key);
            List cList = (List)cMap.get(key);
            if (eList == null) {
                surplus.addAll(cList);
            } else if (cList == null) {
                missing.addAll(eList);
            } else if (eList.size()==1 && cList.size()==1) {
                check(eList, cList);
            } else if (col < columnBindings.length) {
                match(eList, cList, col+1);
            } else {
                surplus.add(cList); surplus.addAll(eList);
            }
        }
    }

    List list (Parse rows) {
        List result = new LinkedList();
        while (rows != null) {
            result.add(rows);
            rows = rows.more;
        }
        return result;
    }

    List list (Object[] rows) {
        List result = new LinkedList();
        for (int i=0; i<rows.length; i++) {
            result.add(rows[i]);
        }
        return result;
    }

    Map eSort(List list, int col) {
        TypeAdapter a = columnBindings[col];
        Map result = new HashMap(list.size());
        for (Iterator i=list.iterator(); i.hasNext(); ) {
            Parse row = (Parse) i.next();
            Parse cell = row.parts.at(col);
            try {
                Object key = a.parse(cell.text());
                bin(result, key, row);
            } catch (Exception e) {
                exception(cell, e);
            }
        }
        return result;
    }

    Map cSort(List list, int col) {
        TypeAdapter a = columnBindings[col];
        Map result = new HashMap(list.size());
        for (Iterator i=list.iterator(); i.hasNext(); ) {
            Object row = i.next();
            try {
                a.target = row;
                Object key = a.get();
                bin(result, key, row);
            } catch (Exception e) {
                // surplus anything with bad keys, including null
                surplus.add(row);
            }
        }
        return result;
    }

    void bin (Map map, Object key, Object row) {
        if (map.containsKey(key)) {
            ((List)map.get(key)).add(row);
        } else {
            List list = new LinkedList();
            list.add(row);
            map.put(key, list);
        }
    }

    Set union (Set a, Set b) {
        Set result = new HashSet();
        result.addAll(a);
        result.addAll(b);
        return result;
    }

    void check (List eList, List cList) {
        // called only in case where both sizes equal one
        Parse row = (Parse)eList.get(0);
        Parse cell = row.parts;
        Object obj = cList.get(0);
        for (int i=0; i<columnBindings.length; i++) {
            TypeAdapter a = columnBindings[i];
            a.target = obj;
            check(cell, a);
            cell = cell.more;
        }
    }

    void mark(Parse rows, String message) {
        String annotation = label(message);
        while (rows != null) {
            wrong(rows.parts);
            rows.parts.addToBody(annotation);
            rows = rows.more;
        }
    }

    void mark(Iterator rows, String message) {
        String annotation = label(message);
        while (rows.hasNext()) {;
            Parse row = (Parse)rows.next();
            wrong(row.parts);
            row.parts.addToBody(annotation);
        }
    }

    Parse buildRows(Object[] rows) {
        Parse root = new Parse(null ,null, null, null);
        Parse next = root;
        for (int i=0; i<rows.length; i++) {
            next = next.more = new Parse("tr", null, buildCells(rows[i]), null);
        }
        return root.more;
    }

    Parse buildCells(Object row) {
        if (row == null) {
            Parse nil = new Parse("td", "null", null, null);
            nil.addToTag(" colspan="+columnBindings.length);
            return nil;
        }
        Parse root = new Parse(null, null, null, null);
        Parse next = root;
        for (int i=0; i<columnBindings.length; i++) {
            next = next.more = new Parse("td", "&nbsp;", null, null);
            TypeAdapter a = columnBindings[i];
            if (a == null) {
                ignore (next);
            } else {
                try {
                    a.target = row;
                    next.body = a.toString(a.get());
                } catch (Exception e) {
                    exception(next, e);
                }
            }
        }
        return root.more;
    }
}
